import logging

from tonie_api.api import TonieAPI
from tonie_api.models import CreativeTonie

from ..models.track_metadata import TrackMetadata
from ..utils import format_seconds


class TonieClient:
    """A client for interacting with the Tonie API."""

    def __init__(self, email: str, password: str, household: str):
        self.api = TonieAPI(email, password)
        self.household = self._get_household(household)

    def _get_household(self, household_name: str):
        """Get the household object by its name.

        Args:
            household_name: The name of the household to retrieve.

        Returns:
            Household: The household object that matches the given name.

        Raises:
            ValueError: If no household with the specified name is found.

        """
        household = next((x for x in self.api.get_households() if x.name == household_name), None)
        if household is None:
            raise ValueError(f"Household [ '{household_name}' ] not found!")
        return household

    def update_creative_tonie(self, creative_tonie_name: str, tracks_metadata: list[TrackMetadata]):
        """Update a creative tonie with new tracks."""
        creative_tonie = self.get_creative_tonie_by_name(creative_tonie_name)
        track_file_names = [str(track.filename) for track in tracks_metadata]
        missing_chapters = self.remove_orphaned_chapters(creative_tonie, track_file_names)
        missing_chapters_metadata = [track for track in tracks_metadata if track.filename in missing_chapters]
        self.upload_tracks_to_creative_tonie(creative_tonie_name, missing_chapters_metadata)
        self.sort_chapters(creative_tonie_name, tracks_metadata)

    def get_creative_tonie_by_name(self, creative_tonie_name: str) -> CreativeTonie:
        """Retrieve a CreativeTonie object by its name from a specified household.

        Args:
            creative_tonie_name: The name of the CreativeTonie to retrieve.

        Returns:
            CreativeTonie: The CreativeTonie object that matches the given name.

        Raises:
            ValueError: If no CreativeTonie with the specified name is found in the household.

        """
        creative_tonie = next(
            (x for x in self.api.get_all_creative_tonies_by_household(self.household) if x.name == creative_tonie_name),
            None,
        )
        if creative_tonie is None:
            raise ValueError(f"Creative Tonie '{creative_tonie_name}' not found!")
        return creative_tonie

    def remove_orphaned_chapters(self, creative_tonie, track_file_names: list[str]):
        """Remove orphaned chapters from a creative tonie.

        Args:
            creative_tonie: The creative tonie to remove orphaned chapters from.
            track_file_names: The list of track file names to keep.

        Returns:
            list[str]: The list of track file names that are missing on the creative tonie.

        """
        logging.info("Removing orphaned chapters from creative tonie...")
        chapters = creative_tonie.chapters
        chapters_to_keep = [chapter for chapter in chapters if chapter.title in track_file_names]
        if chapters_to_keep:
            logging.info(
                f"Keeping chapters: {', '.join(
                [chapter.title for chapter in chapters_to_keep])}"
            )
            self.api.sort_chapter_of_tonie(creative_tonie, chapters_to_keep)
            chapter_titles = [chapter.title for chapter in chapters]
            missing_chapters = [track for track in track_file_names if track not in chapter_titles]
        else:
            logging.info("No chapters to keep found! Deleting all chapters...")
            self.api.clear_all_chapter_of_tonie(creative_tonie)
            missing_chapters = track_file_names
        return missing_chapters

    def sort_chapters(self, creative_tonie_name, tracks_metadata):
        """Sort the chapters of a creative tonie.

        Note:
            Sorting also removes any chapters that are not in the list of tracks_metadata.

        """
        logging.info("Sorting chapters...")
        chapters_sorted = False
        creative_tonie = self.get_creative_tonie_by_name(creative_tonie_name)
        for num, track_metadata in enumerate(tracks_metadata):
            old_chapter_index = next(
                (i for i, item in enumerate(creative_tonie.chapters) if item.title == track_metadata.name),
                -1,
            )
            if old_chapter_index != -1 and num != old_chapter_index:
                creative_tonie.chapters.insert(num, creative_tonie.chapters.pop(old_chapter_index))
                chapters_sorted = True
        if chapters_sorted:
            logging.info("Sorting changed...")
            self.api.sort_chapter_of_tonie(creative_tonie, creative_tonie.chapters)
        else:
            logging.info("No sorting needed...")

    def upload_tracks_to_creative_tonie(self, creative_tonie_name, tracks_metadata: list[TrackMetadata]):
        """Upload tracks to a creative tonie."""
        logging.info("Uploading new tracks / chapters...")
        for track_metadata in tracks_metadata:
            self.upload_track(creative_tonie_name, track_metadata)

    def upload_track(
        self,
        creative_tonie_name: str,
        track_metadata: TrackMetadata,
    ):
        """Upload a track to a creative tonie."""
        creative_tonie = self.get_creative_tonie_by_name(creative_tonie_name)  # TODO: cache this
        chapters = creative_tonie.chapters
        tonie_seconds_remaining = creative_tonie.secondsRemaining
        found_on_creative_tonie = next((x for x in chapters if x.title == track_metadata.artist_and_name), None)

        if found_on_creative_tonie is not None:
            logging.info(f"... skipping '{track_metadata.artist_and_name}' => already present on creative tonie")
            return

        if tonie_seconds_remaining - track_metadata.duration_seconds <= 0:
            logging.warning(
                f"... skipping {track_metadata.artist_and_name} => Not enough free space! Needed: {format_seconds(track_metadata.duration_seconds)} | Free: {format_seconds(tonie_seconds_remaining)}"
            )
            return

        logging.info(f"... uploading '{track_metadata.artist_and_name}' to creative tonie")
        try:
            self.api.upload_file_to_tonie(
                creative_tonie,
                track_metadata.download_path,
                str(track_metadata.filename),
            )
        except FileNotFoundError:
            logging.error(f"... file not found: {track_metadata.download_path}")
            return
        tonie_seconds_remaining = tonie_seconds_remaining - track_metadata.duration_seconds
        logging.info(f"... upload complete! => {format_seconds(tonie_seconds_remaining)} free")



import logging
import os
import sys

from .sync_service import SyncService

logging.basicConfig(
    stream=sys.stdout,
    level=logging.INFO,
    format="%(asctime)s | %(message)s",
)


def main(query: str, creative_tonie_name: str):
    """Sync the files to the creative tonie based on the configuration."""
    sync_service = SyncService()
    sync_service.sync(query=query, creative_tonie_name=creative_tonie_name)


if __name__ == "__main__":
    query = os.environ.get("QUERY")
    creative_tonie_name = os.environ.get("CREATIVE_TONIE")

    if not query or not creative_tonie_name:
        print("Please provide a query and a creative tonie name.")
        sys.exit(1)

    main(query=query, creative_tonie_name=creative_tonie_name)



# from .spotify_client import SpotifyClient
from .spotdl_client import SpotDLClient, SpotDLClientFactory
from .tonie_client import TonieClient
from .youtube_client import YoutubeClient

__all__ = ["TonieClient", "SpotDLClient", "SpotDLClientFactory", "YoutubeClient"]



from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from tonie_sync.clients.spotdl_client import SpotDLClientFactory


@pytest.fixture
def spotdl_client():
    """Return a SpotDLClient instance for testing."""
    with patch("spotdl.Spotdl", return_value=MagicMock()):
        return SpotDLClientFactory.create(
            client_id="test_id", client_secret="test_secret", target_directory=Path("test_music")
        )


def test_sanitize_query_with_valid_url(spotdl_client):
    """Test that the sanitize_query method returns the correct result for a valid URL."""
    url = "https://open.spotify.com/track/7ouMYWpwJ422jRcDASZB7P?si=abc123"
    expected_result = "https://open.spotify.com/track/7ouMYWpwJ422jRcDASZB7P"
    result = spotdl_client.sanitize_query(url)
    assert result == expected_result


def test_sanitize_query_with_valid_uri(spotdl_client):
    """Test that the sanitize_query method returns the correct result for a valid URI."""
    uri = "spotify:track:7ouMYWpwJ422jRcDASZB7P"
    expected_result = "https://open.spotify.com/track/7ouMYWpwJ422jRcDASZB7P"
    result = spotdl_client.sanitize_query(uri)
    assert result == expected_result


def test_sanitize_query_with_invalid_input(spotdl_client):
    """Test that the sanitize_query method raises a ValueError if the input is invalid."""
    invalid_input = "invalid_input"
    with pytest.raises(ValueError):
        spotdl_client.sanitize_query(invalid_input)



import os
from pathlib import Path

import pytubefix
from pydantic import BaseModel, ConfigDict, field_validator
from spotdl.types.song import Song
from spotdl.utils.formatter import create_file_name

from ..utils import fix_filename


class TrackMetadata(BaseModel):
    """Metadata for a track downloaded from a music service."""

    artists: list[str]
    download_root: Path
    name: str
    duration_ms: int
    album_name: str | None = None
    image_url: str | None = None
    release_year: int | None = None
    disc_number: int | None = None
    track_number: int | None = None
    id: str | None = None
    is_playable: bool | None = None

    @field_validator("download_root")
    def validate_download_root(cls, v):
        """Validate the download root path."""
        return Path(os.path.abspath(v))

    @property
    def artist_and_name(self) -> str:
        """Return the artist and track name as a string."""
        return f"{self.artists[0]} - {self.name}"

    @property
    def duration_seconds(self) -> int:
        """Return the duration of the track in seconds."""
        if self.duration_ms is None:
            return 0
        return self.duration_ms // 1000

    @property
    def filename(self) -> str:
        """Return the filename for the downloaded track."""
        filename = fix_filename(f"{self.artist_and_name}.mp3")
        return filename

    @property
    def download_path(self) -> str:
        """Return the download path for the track."""
        return os.path.join(self.download_root, self.filename)


class YoutubeTrackMetadata(TrackMetadata):
    """Metadata for a track downloaded from Youtube."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    youtube_stream: pytubefix.streams.Stream


class SpotifyTrackMetadata(TrackMetadata):
    """Metadata for a track downloaded from Spotify."""

    pass


class SpotDLTrackMetadata(TrackMetadata):
    """Metadata for a track downloaded using SpotDL."""

    song: Song

    @property
    def filename(self) -> str:
        """Return the filename for the downloaded track.

        Note:
        ----
            The filename is generated using the `create_file_name` function from SpotDL.

        """
        return str(
            create_file_name(
                self.song,
                "{artists} - {title}.{output-ext}",
                "mp3",
                restrict=False,
                short=False,
            )
        )



import logging
import os
import re

from tonie_sync.clients.youtube_client import YoutubeClient

from .clients import SpotDLClientFactory, TonieClient


class SyncService:
    """Responsible for syncing music from Spotify to a Toniebox."""

    def __init__(
        self,
        spotify_client_id: str | None = None,
        spotify_client_secret: str | None = None,
        tonie_username: str | None = None,
        tonie_password: str | None = None,
        tonie_household: str | None = None,
        target_directory: str = "./downloads/tonie_sync",
    ) -> None:
        self.spotify_client_id = os.environ.get("SPOTIFY_CLIENT_ID", spotify_client_id)
        self.spotify_client_secret = os.environ.get("SPOTIFY_CLIENT_SECRET", spotify_client_secret)
        self.tonie_username = os.environ.get("TONIE_USERNAME", tonie_username)
        self.tonie_password = os.environ.get("TONIE_PASSWORD", tonie_password)
        self.tonie_household = os.environ.get("TONIE_HOUSEHOLD", tonie_household)
        self.target_directory = os.environ.get("TARGET_DIRECTORY", target_directory)
        if not self.target_directory:
            self.target_directory = os.path.join(os.path.expanduser("~"), "Downloads", "tonie_sync")

    def sync(self, creative_tonie_name: str, query: str):
        """Sync music from Spotify to a Toniebox.

        Args:
        ----
            creative_tonie_name: The name of the creative Tonie.
            query: The search query to find songs on Spotify or Youtube.

        """
        tonie_client = TonieClient(
            email=self.tonie_username,  # type: ignore
            password=self.tonie_password,  # type: ignore
            household=self.tonie_household,  # type: ignore
        )

        if not os.path.exists(self.target_directory):
            logging.info(f"Creating download folder {self.target_directory}")
            os.makedirs(self.target_directory)

        is_youtube_query = re.match(r"(https?://)?(www\.)?(youtube\.com|youtu\.be)/", query)
        if is_youtube_query:
            youtube_client = YoutubeClient(
                target_directory=self.target_directory,  # type: ignore
            )
            logging.info("Downloading from YouTube")
            download_tracks_metadata = youtube_client.search_and_download(query=query)
        else:
            logging.info("Downloading from Spotify")
            spotify_client = SpotDLClientFactory.create(
                client_id=self.spotify_client_id,  # type: ignore
                client_secret=self.spotify_client_secret,  # type: ignore
                target_directory=self.target_directory,  # type: ignore
            )
            download_tracks_metadata = spotify_client.search_and_download(query=query)

        tonie_client.update_creative_tonie(creative_tonie_name, download_tracks_metadata)



# SpotifyClient is removed from the library for now, because authentication
# using username + password does no longer work and zeroconf is not implemented yet.
# import logging
# import os
# import re
# import tempfile
# import time
# from pathlib import Path

# import ffmpy
# import music_tag
# import requests
# from librespot.audio.decoders import AudioQuality, VorbisOnlyAudioQuality
# from librespot.core import Session
# from librespot.metadata import TrackId

# from ..models import SpotifyTrackMetadata
# from ..models.track_metadata import TrackMetadata
# from ..utils import format_seconds


# class SpotifyClient:
#     PLAYLISTS_URL = "https://api.spotify.com/v1/playlists"
#     TRACKS_URL = "https://api.spotify.com/v1/tracks"
#     EPISODE_INFO_URL = "https://api.spotify.com/v1/episodes"
#     SHOWS_URL = "https://api.spotify.com/v1/shows"

#     def __init__(self, username, password, target_directory):
#         self.username = username
#         self.password = password
#         self.session = self._get_spotify_session(target_directory)
#         self._logger = logging.getLogger(__name__)
#         self.download_root = os.path.join(target_directory, "download")
#         os.makedirs(self.download_root, exist_ok=True)

#     def download_tracks(self, download_root: str, download_url: str) -> list[SpotifyTrackMetadata]:
#         """Download all tracks from a Spotify URL.

#         Args:
#             download_root: The path to the data directory.
#             download_url: The URL to download from.

#         Returns:
#             A list of metadata for the downloaded tracks.

#         """
#         url_type, spotify_id = self.parse_url(download_url)
#         if url_type == "playlist":
#             tracks_metadata = self.get_playlist_tracks_metadata(spotify_id)
#         return tracks_metadata

#     def _get_spotify_session(self, target_directory):
#         """Initialize the Spotify session.

#         Args:
#             target_directory: The path to the data directory.

#         """
#         cred_location = os.path.join(target_directory, "credentials.json")
#         if Path(cred_location).is_file():
#             try:
#                 conf = Session.Configuration.Builder().set_store_credentials(False).build()
#                 return Session.Builder(conf).stored_file(cred_location).create()
#             except RuntimeError:
#                 pass
#         else:
#             conf = Session.Configuration.Builder().set_stored_credential_file(cred_location).build()
#             return Session.Builder(conf).user_pass(self.username, self.password).create()

#     def _get_auth_header(self):
#         """Get the authorization header for the Spotify API.

#         Returns:
#             dict: The authorization header.

#         """
#         token = self.session.tokens().get_token(
#             "user-read-email",
#             "playlist-read-private",
#             "user-library-read",
#             "user-follow-read",
#         )
#         header = {
#             "Authorization": f"Bearer {token.access_token}",
#             "Content-Type": "application/json",
#             "app-platform": "WebPlayer",
#         }
#         return header

#     def _make_request(self, url, params=None):
#         """Make a request to the Spotify API.

#         Args:
#             url: The URL to make the request to.
#             params: The parameters to include in the request.

#         Returns:
#             dict: The response from the request.

#         Raises:
#             ValueError: If the response from the URL is invalid or if parsing the response fails.

#         """
#         header = self._get_auth_header()
#         response = requests.get(url, headers=header, params=params)
#         response.raise_for_status()
#         return response.json()

#     def parse_url(self, search_input: str) -> tuple[str, str]:
#         """Parse a Spotify URL or URI and return the ID and object type.

#         Args:
#             search_input: The Spotify URL or URI to parse.

#         Returns:
#             dict[str, str]: A dict with the Id and type of the object.

#         Example:
#             >>> regex_input_for_urls('https://open.spotify.com/track/7ouMYWpwJ422jRcDASZB7P?si=abc123')
#             >>> "track", "7ouMYWpwJ422jRcDASZB7P"

#         Raises:
#             ValueError: If no ID is found in the input.

#         """
#         # Regular expression to match Spotify URLs and URIs
#         pattern = re.compile(
#             r"(?:spotify:(?P<type>\w+):(?P<id>\w+))|(?:https://open.spotify.com/(?P<type2>\w+)/(?P<id2>\w+))"
#         )
#         match = pattern.search(search_input)

#         if match:
#             object_type = match.group("type") or match.group("type2")
#             object_id = match.group("id") or match.group("id2")
#             return object_type, object_id

#         raise ValueError("No ID found in the input.")

#     def get_playlist_tracks_metadata(self, playlist_id: str) -> list[SpotifyTrackMetadata]:
#         """Retrieve all tracks from a specific Spotify playlist.

#         This function fetches all tracks from a playlist identified by its ID from Spotify.
#         It handles pagination to retrieve all tracks by making multiple requests if necessary.

#         Args:
#             playlist_id: The ID of the playlist as a string.

#         Returns:
#             A list of tracks (list[dict]), where each track is represented as a dictionary containing its metadata.

#         Raises:
#             ValueError: If the response from the URL is invalid or if parsing the response fails.

#         """
#         tracks = []
#         offset = 0
#         limit = 100

#         while True:
#             resp = self._make_request(
#                 url=f"{self.PLAYLISTS_URL}/{playlist_id}/tracks",
#                 params={"limit": limit, "offset": offset},
#             )
#             offset += limit
#             tracks.extend(resp["items"])
#             if len(resp["items"]) < limit:
#                 break
#         tracks_metadata = []
#         for track in tracks:
#             track_metadata = self.get_track_metadata(track["track"]["id"])
#             tracks_metadata.append(track_metadata)
#         return tracks_metadata

#     def get_track_metadata(self, track_id) -> SpotifyTrackMetadata:
#         """Retrieve metadata for a specific Spotify track.

#         Args:
#             track_id: The ID of the track as a string.

#         Returns:
#             dict: Metadata for the track.

#         Raises:
#             ValueError: If the response from the URL is invalid or if parsing the response fails.

#         """
#         try:
#             response = self._make_request(
#                 url=f"{self.TRACKS_URL}",
#                 params={"ids": track_id, "market": "from_token"},
#             )
#             track = response["tracks"][0]
#             artists = []
#             for data in track["artists"]:
#                 artists.append(data["name"])
#             track_metadata = SpotifyTrackMetadata(
#                 download_root=self.download_root,
#                 artists=artists,
#                 album_name=track["album"]["name"],
#                 name=track["name"],
#                 image_url=track["album"]["images"][0]["url"],
#                 release_year=track["album"]["release_date"].split("-")[0],
#                 disc_number=track["disc_number"],
#                 track_number=track["track_number"],
#                 id=track["id"],
#                 is_playable=track["is_playable"],
#                 duration_ms=track["duration_ms"],
#             )
#         except Exception as e:
#             raise ValueError(f"Failed to parse TRACKS_URL [ '{track_id}' ] response: {str(e)}") from e
#         return track_metadata

#     def process_track(self, track_metadata: SpotifyTrackMetadata) -> SpotifyTrackMetadata:
#         """Process a track from Spotify.

#         Determine the target location and download the track to the target location, if
#         it does not already exist and is playable.

#         Args:
#             track_metadata: The metadata for the track to process.

#         Returns:
#             SpotifyTrackMetadata: The metadata of the downloaded track.

#         """
#         try:
#             logging.info(f"Processing [ '{track_metadata.name}' ] ...")
#             if os.path.exists(track_metadata.download_path):
#                 self._logger.info("... track already exists => Skipping download.")
#                 return track_metadata

#             if not track_metadata.is_playable:
#                 self._logger.warning("... track is not playable => Skipping download.")
#                 return track_metadata
#             self.download_track(track_metadata)
#             return track_metadata

#         except Exception as e:
#             self._logger.error(f"... failed to download track [ '{track_metadata.name}' ]: {str(e)}")
#             return track_metadata

#     def download_track(
#         self,
#         track_metadata: SpotifyTrackMetadata,
#     ):
#         """Download a track from Spotify."""
#         self._logger.info("... downloading")
#         download_tempfile = tempfile.NamedTemporaryFile(delete=False)
#         librespot_track = TrackId.from_base62(track_metadata.id)
#         stream = self.session.content_feeder().load(
#             librespot_track, VorbisOnlyAudioQuality(AudioQuality.HIGH), False, None
#         )
#         total_size = stream.input_stream.size
#         time_start = time.time()
#         downloaded = 0
#         b = 0
#         while b < 5:
#             data = stream.input_stream.stream().read(20000)
#             download_tempfile.write(data)
#             downloaded += len(data)
#             b += 1 if data == b"" else 0
#             if True:  # if args.ban_protection:
#                 delta_real = time.time() - time_start
#                 delta_want = (downloaded / total_size) * (track_metadata.duration_ms / 5000)
#                 if delta_want > delta_real:
#                     time.sleep(delta_want - delta_real)
#         time_downloaded = time.time()
#         logging.info(f"... finished download in {format_seconds(int(time_downloaded - time_start))}!")
#         self._convert_to_mp3(download_tempfile.name, track_metadata.download_path)
#         self._set_metadata(track_metadata.download_path, track_metadata)

#     def _set_metadata(self, file_path, track_metadata: TrackMetadata) -> None:
#         tags = music_tag.load_file(file_path)
#         tags["albumartist"] = track_metadata.artists[0]
#         tags["artist"] = ", ".join(track_metadata.artists)
#         tags["tracktitle"] = track_metadata.name
#         tags["album"] = track_metadata.album_name
#         tags["year"] = track_metadata.release_year
#         tags["discnumber"] = track_metadata.disc_number
#         tags["tracknumber"] = track_metadata.track_number
#         try:
#             if track_metadata.image_url:
#                 image_url = track_metadata.image_url
#                 img = requests.get(image_url).content
#                 tags["artwork"] = img
#         except Exception as e:
#             logging.error(f"Failed to download image: {str(e)}")
#         tags.save()
#         return

#     def _convert_to_mp3(self, temp_filename, filename) -> None:
#         """Convert raw audio into a playable file.

#         This function uses ffmpy to convert a raw audio file into a playable format.
#         It sets the codec to 'libmp3lame' and the bitrate to '160k' unless the codec is 'copy'.
#         The function handles errors related to the ffmpeg executable not being found.

#         Args:
#             temp_filename: The path to the temporary raw audio file as a string.
#             filename: The path to the output playable audio file as a string.

#         Returns:
#             None

#         """
#         self._logger.info("... converting Track Audio Format")

#         output_params = ["-c:a", "libmp3lame", "-b:a", "160k"]

#         try:
#             ff_m = ffmpy.FFmpeg(
#                 global_options=["-y", "-hide_banner", "-loglevel error"],
#                 inputs={temp_filename: None},
#                 outputs={filename: output_params},
#             )
#             ff_m.run()

#         except ffmpy.FFExecutableNotFoundError:
#             logging.warning("Skipping conversion - ffmpeg not found!")
#         except ffmpy.FFRuntimeError as e:
#             logging.error(f"Failed to convert file: {str(e)}")



import logging
import os
import re
from pathlib import Path

from pytubefix import Playlist, YouTube

from ..models import YoutubeTrackMetadata


class YoutubeClient:
    """Wrapper for the Youtube library pytubefix."""

    _spotdl = None

    def __init__(
        self,
        target_directory: Path = Path("media"),
    ):
        if not os.path.exists(target_directory):
            os.makedirs(target_directory)
        self.target_directory = target_directory
        self._logger = logging.getLogger(__name__)

    def search_and_download(self, query: list[str] | str) -> list[YoutubeTrackMetadata]:
        """Search for songs on Youtube and download them.

        Args:
        ----
            query: The search query to find songs on Youtube.

        Returns:
        -------
            list[YoutubeTrackMetadata]: A list of YoutubeTrackMetadata objects for the downloaded songs.

        Example:
        -------
            >>> client.search_and_download("https://youtube.com/playlist?list=OLAK5uy_kP41f7I4coTZ1yh9G_FyQrJSFiS4eQQgw&si=wgQvRwuCBGgP5-Io")

        """
        if isinstance(query, str):
            queries = [query]
        else:
            queries = queries
        for query in queries:
            is_playlist = re.match(r"^.*\.(com|be)/playlist\?list=.*$", query)
            tracks_metadata: list[YoutubeTrackMetadata] = []
            if is_playlist:
                pl = Playlist(query)
                for track in pl.videos:
                    track_metadata = YoutubeTrackMetadata(
                        artists=[track.author],
                        download_root=self.target_directory,
                        name=track.title,
                        duration_ms=track.length * 1000,  # convert to ms
                        album_name=None,
                        image_url=track.thumbnail_url,
                        release_year=track.publish_date.year if track.publish_date else None,
                        disc_number=None,
                        track_number=None,
                        id=track.video_id,
                        is_playable=True,  # always True for Youtube songs,
                        youtube_stream=track.streams.get_audio_only(),
                    )
                    tracks_metadata.append(track_metadata)
            else:
                track = YouTube(query)
                track_metadata = YoutubeTrackMetadata(
                    artists=[track.author],
                    download_root=self.target_directory,
                    name=track.title,
                    duration_ms=track.length * 1000,  # convert to ms
                    album_name=None,
                    image_url=track.thumbnail_url,
                    release_year=track.publish_date.year if track.publish_date else None,
                    disc_number=None,
                    track_number=None,
                    id=track.video_id,
                    is_playable=True,  # always True for Youtube songs
                    youtube_stream=track.streams.get_audio_only(),
                )
                tracks_metadata.append(track_metadata)
        for track_metadata in tracks_metadata:
            self._logger.info(f"Downloading {track.title} from Youtube")
            track_metadata.youtube_stream.download(
                output_path=track_metadata.download_root,
                filename=track_metadata.filename,
            )
        return tracks_metadata
